/*!
 * @vueblocks/vue-use-core v0.2.2
 * (c) 2021 xiaoluoboding
 * @license MIT
 */
import { customRef, ref, watch, getCurrentInstance, onMounted, nextTick, onUnmounted, reactive } from 'vue-demi';
import debounce from 'lodash.debounce';
import { ResizeObserver } from '@juggle/resize-observer';
import screenfull from 'screenfull';
import throttle from 'lodash.throttle';
import keymaster from 'keymaster';

const useDebouncedRef = (value, delay = 200) => {
    return customRef((track, trigger) => {
        return {
            get() {
                track();
                return value;
            },
            set: debounce((newValue) => {
                value = newValue;
                trigger();
            }, delay)
        };
    });
};

/**
 * It immediately detects when an element resizes and provides accurate sizing information back to the handler.
 * @param target
 * @param callback
 * @param options
 */
const useResizeObserver = (target, callback, options) => {
    let ro;
    const targetRef = ref(target);
    const isSupported = window && 'ResizeObserver' in window;
    const stopWatch = watch(targetRef, (newVal) => {
        if (isSupported && window && newVal) {
            ro = new ResizeObserver(callback);
            ro.observe(newVal, options);
        }
    }, { immediate: true });
    const stop = () => {
        ro && ro.disconnect();
        ro = undefined;
        stopWatch();
    };
    return {
        stop
    };
};

const isBoolean = (val) => typeof val === 'boolean';
const isClient = ![typeof window, typeof document].includes('undefined');
const isElement = (val) => typeof val === 'object' && !!val.tagName;

const useElementSize = (target, options) => {
    const width = ref(isElement(target) ? target.clientWidth : 0);
    const height = ref(isElement(target) ? target.clientHeight : 0);
    useResizeObserver(target, ([entry]) => {
        width.value = entry.contentRect.width;
        height.value = entry.contentRect.height;
    }, options);
    return { width, height };
};

/**
 * Call onMounted() if it's inside a component lifecycle, if not, run just call the function
 * @param fn
 * @param sync if set to false, it will run in the nextTick() of Vue
 */
function tryOnMounted(fn, sync = true) {
    if (getCurrentInstance()) {
        onMounted(fn);
    }
    else if (sync) {
        fn();
    }
    else {
        nextTick(fn);
    }
}
/**
 * Call onUnmounted() if it's inside a component lifecycle, if not, do nothing
 * @param fn
 */
function tryOnUnmounted(fn) {
    if (getCurrentInstance())
        onUnmounted(fn);
}

/**
 * Reactive Fullscreen API build on top of screenfull.js
 * @param {HTMLElement} target
 * @return {boolean} isFullscreen
 * @return {function} toggleFullscreen
 */
const useFullscreen = (target) => {
    const targetRef = ref(target || document.querySelector('html'));
    const isFullscreen = ref(false);
    const toggleFullscreen = () => {
        screenfull.isEnabled && screenfull.toggle(targetRef.value);
    };
    tryOnMounted(() => {
        if (screenfull.isEnabled) {
            screenfull.on('change', () => {
                isFullscreen.value = screenfull.isFullscreen;
            });
        }
    });
    return {
        isFullscreen,
        toggleFullscreen
    };
};

function useReactiveState(getter) {
    const state = reactive(getter());
    const update = () => {
        Object.assign(state, getter());
    };
    return [state, update];
}

/**
 * Reactive Resize API, When window resize, do someting.
 * @param {function} fn
 * @param {object} options
 * @return {function} dispatchResize
 */
const useResizeEvent = (fn, options = { useCapture: false, useThrottle: null, delay: 500 }) => {
    const { useCapture = false, useThrottle = null, delay = 500 } = options;
    const dispatchResize = () => {
        const e = document.createEvent('Event');
        e.initEvent('resize', true, true);
        window.dispatchEvent(e);
    };
    const waitFn = isBoolean(useThrottle)
        ? useThrottle
            ? throttle(fn, delay)
            : debounce(fn, delay)
        : fn;
    tryOnMounted(() => {
        window.addEventListener('resize', waitFn, useCapture);
    });
    tryOnUnmounted(() => {
        window.removeEventListener('resize', waitFn, useCapture);
    });
    return {
        dispatchResize
    };
};

const bindKeyHandler = (fn) => {
    return (...args) => {
        fn(...args);
        return false;
    };
};
/**
 * Given shortcut a new bind function
 */
const shortcut = {
    bind: (seed, func) => keymaster(seed, bindKeyHandler(func)),
    ...keymaster
};
/**
 * Bind shortcut onMounted & unbind shortcut automatically onUnmounted in setup hooks
 * or
 * Unbind shortcut before bind shortcut without setup hooks
 * @param {Object} keymap
 * @param {Object} vm
 * @return {function} toggle
 */
const useShortcut = (keymap, vm) => {
    let sync = true;
    if (!vm) {
        vm = getCurrentInstance();
    }
    else {
        sync = false;
        unbindKeys();
    }
    function bindKeys() {
        Object.keys(keymap).forEach(key => {
            shortcut.bind(key, keymap[key]);
        });
    }
    function unbindKeys() {
        Object.keys(keymap).forEach(key => {
            shortcut.unbind(key);
        });
    }
    tryOnMounted(() => bindKeys(), sync);
    tryOnUnmounted(unbindKeys);
    return shortcut;
};

/**
 * Reactive window size
 * @param {object} options
 * @return {Ref<number>} WindowSize
 */
const useWindowSize = (options) => {
    const width = ref(isClient ? window.innerWidth : 0);
    const height = ref(isClient ? window.innerHeight : 0);
    const onResize = () => {
        width.value = window.innerWidth;
        height.value = window.innerHeight;
    };
    const { dispatchResize } = useResizeEvent(onResize, options);
    onMounted(dispatchResize);
    return { width, height };
};

export { tryOnMounted, tryOnUnmounted, useDebouncedRef, useElementSize, useFullscreen, useReactiveState, useResizeEvent, useResizeObserver, useShortcut, useWindowSize };
